const std = @import("std");
const allegro = @import("allegro");
const engine = @import("engine.zig");
const paths = @import("paths.zig");
const Palette = @import("palette.zig").Palette;
const Scene = @import("scene.zig").Scene;
const Renderer = @import("renderer.zig").Renderer;

pub const Context = struct {
    const DefaultDisplayWidth = 960;
    const DefaultDisplayHeight = 540;
    // const DefaultDisplayWidth = 480;
    // const DefaultDisplayHeight = 270;

    allocator: std.mem.Allocator,
    palette: Palette = undefined,
    shouldQuit: bool = false,
    showFPS: bool = false,
    showDebug: bool = false,
    scene: ?Scene = null,

    events: allegro.EventQueue,
    renderer: Renderer,

    fps: engine.FPS,
    keys: engine.Keys(c_int),

    pub fn init(allocator: std.mem.Allocator) !Context {
        if (!allegro.init()) {
            return error.FailedToInitialize;
        }
        if (!allegro.initImageAddon()) {
            return error.FailedToInitialize;
        }
        if (!allegro.installKeyboard()) {
            return error.FailedToInitialize;
        }
        if (!allegro.installMouse()) {
            return error.FailedToInitialize;
        }
        if (!allegro.initFontAddon()) {
            return error.FailedToInitialize;
        }
        if (!allegro.initTtfAddon()) {
            return error.FailedToInitialize;
        }
        const events = try allegro.createEventQueue();

        allegro.setNewDisplayOption(allegro.NewDisplayOption.VSYNC, 1, allegro.OptionImportance.SUGGEST);
        // allegro.setNewDisplayOption(allegro.NewDisplayOption.SAMPLE_BUFFERS, 1, allegro.OptionImportance.REQUIRE);
        // allegro.setNewDisplayOption(allegro.NewDisplayOption.SAMPLES, 4, allegro.OptionImportance.REQUIRE);
        allegro.setNewDisplayFlags(allegro.NewDisplayFlags{ .RESIZABLE = true });
        const display = try allegro.createDisplay(DefaultDisplayWidth, DefaultDisplayHeight);

        events.registerDisplay(display);
        events.registerKeyboard();
        events.registerMouse();

        return Context{
            .allocator = allocator,
            .events = events,
            .renderer = Renderer.init(allocator, display),
            .fps = engine.FPS{},
            .keys = engine.Keys(c_int).init(allocator),
        };
    }

    fn exitScene(self: *Context) void {
        if (self.scene) |scene| {
            scene.exit(self);
            scene.deinit();
            self.scene = null;
        }
    }

    pub fn deinit(self: *Context) void {
        self.exitScene();
        self.renderer.deinit();
        self.events.destroy();
    }

    pub fn registerFonts(self: *Context) !void {
        const viewport = self.renderer.viewport;
        const fonts = &self.renderer.fonts;
        try fonts.addFromFileTTF("default", paths.AssetsDir ++ "/fonts/Pixellari.ttf", viewport.scaledInteger(32));
        try fonts.addFromFileTTF("small", paths.AssetsDir ++ "/fonts/Pixellari.ttf", viewport.scaledInteger(16));
        try fonts.addFromFileTTF("extra-small", paths.AssetsDir ++ "/fonts/Pixellari.ttf", viewport.scaledInteger(12));
        try fonts.addFromFileTTF("large", paths.AssetsDir ++ "/fonts/Pixellari.ttf", viewport.scaledInteger(64));
        try fonts.addFromFileTTF("debug", paths.AssetsDir ++ "/fonts/Cabin-Regular.ttf", viewport.scaledInteger(16));
    }

    pub fn quit(self: *Context) void {
        self.shouldQuit = true;
    }

    pub fn switchToScene(self: *Context, comptime SceneType: type, build: ?*const fn (*SceneType) void) !void {
        self.exitScene();

        const scene = try Scene.init(SceneType, self.allocator, build);
        self.scene = scene;
        scene.enter(self);
    }
};
