const std = @import("std");
const allegro = @import("allegro");

pub fn Assets(comptime Asset: type, comptime destroy: *const fn (Asset) void) type {
    return struct {
        const Self = @This();

        assets: std.StringHashMap(Asset),

        pub fn init(allocator: std.mem.Allocator) Self {
            return Self{
                .assets = std.StringHashMap(Asset).init(allocator),
            };
        }

        pub fn deinit(self: *Self) void {
            var assets = self.assets.iterator();
            while (assets.next()) |entry| {
                destroy(entry.value_ptr.*);
            }
            self.assets.deinit();
        }

        pub fn get(self: Self, name: []const u8) ?Asset {
            return self.assets.get(name);
        }

        pub fn put(self: *Self, name: []const u8, asset: Asset) !void {
            try self.assets.put(name, asset);
        }
    };
}

pub const Fonts = struct {
    fn destroy(font: allegro.Font) void {
        font.destroy();
    }

    const Map = Assets(allegro.Font, Fonts.destroy);

    assets: Map,

    pub fn init(allocator: std.mem.Allocator) Fonts {
        return Fonts{
            .assets = Map.init(allocator),
        };
    }

    pub fn deinit(self: *Fonts) void {
        self.assets.deinit();
    }

    pub fn addFromFileTTF(self: *Fonts, name: []const u8, path: [*:0]const u8, size: i32) !void {
        const font = try allegro.loadTtfFont(path, size, allegro.LoadTtfFontFlags{});
        if (self.assets.get(name)) |f| {
            f.destroy();
        }
        try self.assets.put(name, font);
    }

    pub fn get(self: Fonts, name: []const u8) ?allegro.Font {
        return self.assets.get(name);
    }
};

pub const Textures = struct {
    fn destroy(bitmap: allegro.Bitmap) void {
        bitmap.destroy();
    }
    const Map = Assets(allegro.Bitmap, destroy);

    assets: Map,

    pub fn init(allocator: std.mem.Allocator) Textures {
        return Textures{
            .assets = Map.init(allocator),
        };
    }

    pub fn deinit(self: *Textures) void {
        self.assets.deinit();
    }

    pub fn addFromFile(self: *Textures, name: []const u8, path: [*:0]const u8) !void {
        const bitmap = try allegro.loadBitmap(path);
        try self.assets.put(name, bitmap);
    }

    pub fn addFromMemoryPNG(self: *Textures, name: []const u8, data: []const u8) !void {
        const memoryFile = try allegro.openMemfile(@constCast(data), "r");
        defer _ = allegro.fclose(memoryFile);

        const bitmap = try allegro.loadBitmapF(memoryFile, ".png");
        try self.assets.put(name, bitmap);
    }

    pub fn get(self: Textures, name: []const u8) ?allegro.Bitmap {
        return self.assets.get(name);
    }
};

pub const Sprite = struct {
    allocator: std.mem.Allocator,
    frames: []allegro.Bitmap,

    pub fn init(allocator: std.mem.Allocator, bitmap: allegro.Bitmap, width: i32, height: i32) !Sprite {
        const bitmapWidth = bitmap.width();
        const bitmapHeight = bitmap.height();

        const horizontal = @divTrunc(bitmapWidth, width);
        const vertical = @divTrunc(bitmapHeight, height);
        const n = @intCast(usize, horizontal * vertical);
        const frames = try allocator.alloc(allegro.Bitmap, n);
        var i: i32 = 0;
        while (i < n) : (i += 1) {
            const left = @mod(i, horizontal) * width;
            const top = @divTrunc(i, horizontal) * height;
            frames[@intCast(usize, i)] = try bitmap.sub(left, top, width, height);
        }

        return Sprite{
            .allocator = allocator,
            .frames = frames,
        };
    }

    pub fn deinit(self: Sprite) void {
        for (self.frames) |f| {
            f.destroy();
        }
        self.allocator.free(self.frames);
    }

    pub fn frame(self: Sprite, i: usize) allegro.Bitmap {
        return self.frames[i];
    }
};

pub const Sprites = struct {
    fn deinitAsset(sprite: Sprite) void {
        sprite.deinit();
    }
    const Map = Assets(Sprite, deinitAsset);

    allocator: std.mem.Allocator,
    assets: Map,

    pub fn init(allocator: std.mem.Allocator) Sprites {
        return Sprites{
            .allocator = allocator,
            .assets = Map.init(allocator),
        };
    }

    pub fn deinit(self: *Sprites) void {
        self.assets.deinit();
    }

    pub fn add(self: *Sprites, name: []const u8, bitmap: allegro.Bitmap, width: i32, height: i32) !void {
        try self.assets.put(name, try Sprite.init(self.allocator, bitmap, width, height));
    }

    pub fn addFromTextures(self: *Sprites, textures: Textures, name: []const u8, width: i32, height: i32) !void {
        try self.add(name, textures.get(name).?, width, height);
    }

    pub fn get(self: Sprites, name: []const u8) ?Sprite {
        return self.assets.get(name);
    }

    pub fn getFrame(self: Sprites, name: []const u8, i: usize) ?allegro.Bitmap {
        if (self.assets.get(name)) |sprite| {
            return sprite.frame(i);
        }
        return null;
    }
};
