const std = @import("std");
const OpaquePtr = @import("opaque_ptr.zig").OpaquePtr;

pub fn Scene(comptime Context: type, comptime Event: type) type {
    return struct {
        const Self = @This();

        object: usize = undefined,
        virtualTable: VirtualTable = undefined,
        opaquePtr: ?OpaquePtr = null,

        const VirtualTable = struct {
            enter: *const fn (Self, *Context) void,
            exit: *const fn (Self, *Context) void,
            handle: *const fn (Self, *Context, Event) anyerror!void,
            tick: *const fn (Self, *Context, f32, f32) void,
            render: *const fn (Self, *Context) void,
        };

        pub fn enter(self: Self, ctx: *Context) void {
            self.virtualTable.enter(self, ctx);
        }

        pub fn exit(self: Self, ctx: *Context) void {
            self.virtualTable.exit(self, ctx);
        }

        pub fn handle(self: Self, ctx: *Context, event: Event) !void {
            try self.virtualTable.handle(self, ctx, event);
        }

        pub fn render(self: Self, ctx: *Context) void {
            self.virtualTable.render(self, ctx);
        }

        pub fn tick(self: Self, ctx: *Context, t: f32, dt: f32) void {
            self.virtualTable.tick(self, ctx, t, dt);
        }

        pub fn makeOpaque(comptime SceneType: type, object: *SceneType) Self {
            return Self{ .object = @ptrToInt(object), .virtualTable = .{
                .enter = struct {
                    fn enter(self: Self, ctx: *Context) void {
                        const scene = @intToPtr(*SceneType, self.object);
                        scene.enter(ctx);
                    }
                }.enter,
                .exit = struct {
                    fn exit(self: Self, ctx: *Context) void {
                        const scene = @intToPtr(*SceneType, self.object);
                        scene.exit(ctx);
                    }
                }.exit,
                .handle = struct {
                    fn handle(self: Self, ctx: *Context, event: Event) !void {
                        const scene = @intToPtr(*SceneType, self.object);
                        try scene.handle(ctx, event);
                    }
                }.handle,
                .tick = struct {
                    fn tick(self: Self, ctx: *Context, t: f32, dt: f32) void {
                        const scene = @intToPtr(*SceneType, self.object);
                        scene.tick(ctx, t, dt);
                    }
                }.tick,
                .render = struct {
                    fn render(self: Self, ctx: *Context) void {
                        const scene = @intToPtr(*SceneType, self.object);
                        scene.render(ctx);
                    }
                }.render,
            } };
        }

        pub fn init(comptime SceneType: type, allocator: std.mem.Allocator, build: ?*const fn (*SceneType) void) !Self {
            const object = try OpaquePtr.create(SceneType, allocator);
            var scene = Self.makeOpaque(SceneType, object.ptr);
            if (build) |b| {
                b(object.ptr);
            }
            return scene;
        }

        pub fn deinit(self: Self) void {
            if (self.opaquePtr) |ptr| {
                ptr.destroy();
            }
        }
    };
}
