const std = @import("std");
const allegro = @import("allegro");
const engine = @import("../engine.zig");
const Animation = @import("animation.zig").Animation;
const Level = @import("level.zig").Level;
const Renderer = @import("../renderer.zig").Renderer;
const TileMap = @import("tile_map.zig").TileMap;

pub const Game = struct {
    pub const Direction = enum {
        Left,
        Right,
    };

    pub const State = enum {
        InProgress,
        Over,
        Completed,
    };

    pub const TileSize = 32;

    prng: std.rand.DefaultPrng = std.rand.DefaultPrng.init(0),

    level: Level,
    state: State = .InProgress,
    starsCollected: usize = 0,
    starsTotal: usize,

    playerPosition: engine.PointF,
    playerFallingAnimation: Animation,
    playerIdleAnimation: Animation,
    playerWalkingAnimation: Animation,
    playerDirection: Direction,
    playerVelocity: engine.PointF,

    starAnimation: Animation,

    renderer: *Renderer,
    randomTile: TileMap(usize),
    collected: TileMap(bool),
    // current viewport translated to tiles.
    boundsInTiles: engine.RectangleF,

    pub fn init(allocator: std.mem.Allocator, level: Level, renderer: *Renderer) Game {
        const playerPosition = level.character.float();
        return Game{
            .level = level,
            .starsTotal = level.collectables.countValue(.Star),
            .playerPosition = playerPosition,
            .playerFallingAnimation = Animation.initPartialLoop(renderer.sprites.get("character_lion_48").?, 0.2, 8, 12),
            .playerIdleAnimation = Animation.initPartialLoop(renderer.sprites.get("character_lion_48").?, 0.25, 0, 4),
            .playerWalkingAnimation = Animation.initPartialLoop(renderer.sprites.get("character_lion_48").?, 0.125, 4, 8),
            .playerDirection = Direction.Right,
            .playerVelocity = engine.PointF.Zero,
            .starAnimation = Animation.init(renderer.sprites.get("item_star_32").?, 0.15),
            .renderer = renderer,
            .randomTile = TileMap(usize).init(allocator),
            .collected = TileMap(bool).init(allocator),
            .boundsInTiles = Game.calculateBoundsInTiles(playerPosition),
        };
    }

    pub fn deinit(self: *Game) void {
        self.level.deinit();
        self.randomTile.deinit();
    }

    fn calculateBoundsInTiles(position: engine.PointF) engine.RectangleF {
        return engine.RectangleF.initRelative(
            position.x - 15,
            @min(position.y - 8.875, 4.125),
            30,
            16.875,
        );
    }

    pub fn drawSpriteFrame(self: *Game, spriteName: []const u8, frame: usize, x: i64, y: i64) void {
        self.drawSpriteFrameP(spriteName, frame, engine.pt(x, y).float());
    }

    pub fn drawSpriteFrameP(self: *Game, spriteName: []const u8, frame: usize, position: engine.PointF) void {
        const view = self.tileP(position);
        self.renderer.drawSpriteFrameV(spriteName, frame, view.x, view.y);
    }

    pub fn debugHighlightTile(self: Game, position: engine.PointF) void {
        const renderer = self.renderer;
        const topLeft = self.tileP(position);
        const topLeftScreen = renderer.viewport.viewToScreenP(topLeft);
        const tileSizeScreen = TileSize * renderer.viewport.scale;
        renderer.textures.get("opaque").?.drawTintedScaled(allegro.mapRgba(127, 127, 127, 127), topLeftScreen.x, topLeftScreen.y, tileSizeScreen, tileSizeScreen);
    }

    pub fn movePlayer(self: *Game, dt: f32) void {
        if (self.state != .InProgress) return;

        const to = self.playerPosition.add(self.playerVelocity.multiply(dt));
        self.playerPosition = to;

        if (self.playerPosition.y > 21.5) {
            self.state = .Over;
            return;
        }

        if (self.playerVelocity.y > 0 and self.playerIsOnTile()) {
            self.playerPosition.y = std.math.floor(self.playerPosition.y);
            self.playerVelocity.y = 0;
        } else if (self.playerVelocity.y < 0 and self.playerIsUnderTile()) {
            self.playerPosition.y = std.math.ceil(self.playerPosition.y);
            self.playerVelocity.y = 0.1;
        }
        if (self.playerVelocity.x != 0) {
            const playerTileTop = @floatToInt(i64, std.math.floor(self.playerPosition.y));
            const playerTileBottom = @floatToInt(i64, std.math.floor(self.playerPosition.y + 0.99));
            if (self.playerVelocity.x < 0) {
                const playerTileLeft = @floatToInt(i64, std.math.ceil(self.playerPosition.x)) - 1;
                if (self.level.tiles.get(playerTileLeft, playerTileTop) != null or self.level.tiles.get(playerTileLeft, playerTileBottom) != null) {
                    self.playerPosition.x = std.math.ceil(self.playerPosition.x);
                    self.playerVelocity.x = 0;
                }
            }
            if (self.playerVelocity.x > 0) {
                const playerTileRight = @floatToInt(i64, std.math.floor(self.playerPosition.x)) + 1;
                if (self.level.tiles.get(playerTileRight, playerTileTop) != null or self.level.tiles.get(playerTileRight, playerTileBottom) != null) {
                    self.playerPosition.x = std.math.floor(self.playerPosition.x);
                    self.playerVelocity.x = 0;
                }
            }
        }
        self.playerDirection = if (self.playerVelocity.x == 0) self.playerDirection else if (self.playerVelocity.x > 0) Direction.Right else Direction.Left;

        const playerCenter = self.playerPosition.add(engine.PointF.Half).floor().integer();
        var dx: i64 = -1;
        while (dx <= 1) : (dx += 1) {
            var dy: i64 = -1;
            while (dy <= 1) : (dy += 1) {
                const x = playerCenter.x + dx;
                const y = playerCenter.y + dy;
                const collectablePosition = engine.pt(x, y);

                if (self.level.collectables.get(x, y)) |collectable| {
                    if (self.collected.hasValue(x, y, true)) continue;
                    if (self.playerPosition.distance2(collectablePosition.float()) > 1.5) continue;

                    switch (collectable) {
                        .Exit => self.state = .Completed,
                        .Star => {
                            self.starsCollected += 1;
                            self.collected.set(x, y, true) catch {};
                        },
                    }
                }
            }
        }

        self.boundsInTiles = Game.calculateBoundsInTiles(self.playerPosition);
    }

    pub fn playerIsOnTile(self: Game) bool {
        const playerTileHeight = @floatToInt(i64, std.math.floor(self.playerPosition.y)) + 1;
        const playerTileLeft = @floatToInt(i64, std.math.floor(self.playerPosition.x + 0.5));
        const playerTileRight = @floatToInt(i64, std.math.ceil(self.playerPosition.x - 0.5));

        return self.level.tiles.get(playerTileLeft, playerTileHeight) != null or self.level.tiles.get(playerTileRight, playerTileHeight) != null;
    }

    pub fn playerIsUnderTile(self: Game) bool {
        const playerTileHeight = @floatToInt(i64, std.math.ceil(self.playerPosition.y)) - 1;
        const playerTileLeft = @floatToInt(i64, std.math.floor(self.playerPosition.x + 0.5));
        const playerTileRight = @floatToInt(i64, std.math.ceil(self.playerPosition.x - 0.5));

        return self.level.tiles.get(playerTileLeft, playerTileHeight) != null or self.level.tiles.get(playerTileRight, playerTileHeight) != null;
    }

    pub fn randomTileOffset(self: *Game, x: i64, y: i64, maxOffset: usize) usize {
        if (self.randomTile.get(x, y)) |offset| {
            return offset;
        }
        const offset = self.prng.random().intRangeLessThan(usize, 0, maxOffset);
        self.randomTile.set(x, y, offset) catch {};
        return offset;
    }

    // return values are in view coordinates
    pub fn tile(self: Game, x: f32, y: f32) engine.PointF {
        return self.tileP(engine.ptF(x, y));
    }

    pub fn tileCenter(self: Game, x: i64, y: i64) engine.PointF {
        return self.tileCenterP(engine.pt(x, y));
    }

    pub fn tileCenterP(self: Game, position: engine.Point) engine.PointF {
        const center = engine.PointF.Half;
        return self.tileP(position.float().add(center));
    }

    pub fn tileP(self: Game, position: engine.PointF) engine.PointF {
        const relative = position.subtract(self.boundsInTiles.min);
        return engine.ptF(relative.x / self.boundsInTiles.width(), relative.y / self.boundsInTiles.height());
    }

    pub fn tileTopLeft(self: Game, x: i64, y: i64) engine.PointF {
        return self.tileTopLeftP(engine.pt(x, y));
    }

    pub fn tileTopLeftP(self: Game, p: engine.Point) engine.PointF {
        return self.tileP(p.float());
    }

    pub fn tilesInView(self: Game) engine.Rectangle {
        // tiles that are partially or completely in view.
        return engine.RectangleF.initAbsolute(
            std.math.floor(self.boundsInTiles.min.x),
            std.math.floor(self.boundsInTiles.min.y),
            std.math.floor(self.boundsInTiles.max.x) + 1,
            std.math.floor(self.boundsInTiles.max.y) + 1,
        ).integer();
    }

    pub fn tick(self: *Game, t: f32, dt: f32) void {
        _ = t;
        _ = self;
        _ = dt;
    }
};
