const std = @import("std");
const Object = @import("../engine/object.zig").Object;

pub fn TileMap(comptime Value: type) type {
    return struct {
        const Self = @This();

        pub const Column = struct {
            values: std.AutoHashMap(i64, Value),

            pub fn init(allocator: std.mem.Allocator) Column {
                return Column{
                    .values = std.AutoHashMap(i64, Value).init(allocator),
                };
            }

            pub fn deinit(self: *Column) void {
                self.values.deinit();
            }

            pub fn countValue(self: Column, value: Value) usize {
                var count: usize = 0;
                var values = self.values.valueIterator();
                while (values.next()) |v| {
                    if (value == v.*) {
                        count += 1;
                    }
                }
                return count;
            }

            pub fn get(self: Column, row: i64) ?Value {
                return self.values.get(row);
            }

            pub fn set(self: *Column, row: i64, value: Value) !void {
                try self.values.put(row, value);
            }
        };

        pub const Ordinals = struct {
            left: bool,
            top: bool,
            right: bool,
            bottom: bool,

            pub fn count(self: Ordinals) usize {
                var n: usize = 0;
                if (self.left) n += 1;
                if (self.top) n += 1;
                if (self.right) n += 1;
                if (self.bottom) n += 1;
                return n;
            }
        };

        allocator: std.mem.Allocator,
        columns: std.AutoHashMap(i64, Column),

        pub fn init(allocator: std.mem.Allocator) Self {
            return Self{
                .allocator = allocator,
                .columns = std.AutoHashMap(i64, Column).init(allocator),
            };
        }

        pub fn deinit(self: *Self) void {
            var columns = self.columns.valueIterator();
            while (columns.next()) |c| {
                c.deinit();
            }
            self.columns.deinit();
        }

        pub fn column(self: Self, x: i64) *Column {
            return self.columns.getPtr(x).?;
        }

        pub fn countValue(self: Self, value: Value) usize {
            var count: usize = 0;
            var columns = self.columns.valueIterator();
            while (columns.next()) |c| {
                count += c.countValue(value);
            }
            return count;
        }

        pub fn ensureColumn(self: *Self, x: i64) !*Column {
            if (self.columns.getPtr(x)) |c| {
                return c;
            }
            try self.columns.put(x, Column.init(self.allocator));
            return self.columns.getPtr(x).?;
        }

        pub fn get(self: Self, x: i64, y: i64) ?Value {
            if (self.columns.get(x)) |c| {
                return c.get(y);
            }
            return null;
        }

        pub fn getOrdinals(self: Self, x: i64, y: i64, value: Value) Ordinals {
            return Ordinals{
                .left = self.hasValue(x - 1, y, value),
                .top = self.hasValue(x, y - 1, value),
                .right = self.hasValue(x + 1, y, value),
                .bottom = self.hasValue(x, y + 1, value),
            };
        }

        pub fn hasValue(self: Self, x: i64, y: i64, value: Value) bool {
            if (self.get(x, y)) |v| {
                if (value == v) {
                    return true;
                }
            }
            return false;
        }

        pub fn set(self: *Self, x: i64, y: i64, value: Value) !void {
            const c = try self.ensureColumn(x);
            try c.set(y, value);
        }
    };
}
