const std = @import("std");
const allegro = @import("allegro");
const engine = @import("engine.zig");
const game = @import("game.zig");
const paths = @import("paths.zig");
const Context = @import("context.zig").Context;
const Renderer = @import("renderer.zig").Renderer;

pub const GameScene = struct {
    game: game.Game = undefined,

    pub fn enter(self: *GameScene, ctx: *Context) void {
        self.load(ctx);
    }

    fn load(self: *GameScene, ctx: *Context) void {
        const level = game.Level.init(ctx.allocator, paths.AssetsDir ++ "/levels/level1.txt") catch unreachable;
        self.game = game.Game.init(ctx.allocator, level, &ctx.renderer);
    }

    pub fn exit(self: *GameScene, ctx: *Context) void {
        self.game.deinit();
        _ = ctx;
    }

    pub fn handle(self: *GameScene, ctx: *Context, event: *allegro.Event) !void {
        if (self.game.state != .InProgress) {
            switch (event.type) {
                allegro.c.ALLEGRO_EVENT_KEY_DOWN => {
                    switch (event.keyboard.keycode) {
                        allegro.c.ALLEGRO_KEY_ESCAPE => {
                            ctx.quit();
                        },
                        allegro.c.ALLEGRO_KEY_ENTER => {
                            self.game.deinit();
                            self.load(ctx);
                        },
                        else => {},
                    }
                },
                else => {},
            }
        }
    }

    pub fn tick(self: *GameScene, ctx: *Context, t: f32, dt: f32) void {
        const horizontalWalkAcceleration: f32 = 13;
        const gravity: f32 = 71;
        const jumpVelocity: f32 = -26;
        const maxFallVelocity: f32 = 23;
        const maxHorizontalWalkVelocity: f32 = 7; // tiles/s

        if (self.game.state != .InProgress) {
            self.game.playerIdleAnimation.tick(t);
            return;
        }

        if (ctx.keys.isKeyPressed(allegro.c.ALLEGRO_KEY_LEFT)) {
            self.game.playerVelocity.x = std.math.max(-maxHorizontalWalkVelocity, self.game.playerVelocity.x - horizontalWalkAcceleration);
        } else if (ctx.keys.isKeyPressed(allegro.c.ALLEGRO_KEY_RIGHT)) {
            self.game.playerVelocity.x = std.math.min(maxHorizontalWalkVelocity, self.game.playerVelocity.x + horizontalWalkAcceleration);
        } else {
            if (std.math.fabs(self.game.playerVelocity.x) > 0.2) {
                self.game.playerVelocity.x = self.game.playerVelocity.x - std.math.sign(self.game.playerVelocity.x) * horizontalWalkAcceleration * dt;
            } else {
                self.game.playerVelocity.x = 0;
            }
        }

        if (self.game.playerVelocity.y == 0 and ctx.keys.isKeyPressed(allegro.c.ALLEGRO_KEY_SPACE)) {
            self.game.playerVelocity.y = jumpVelocity;
        } else if (!self.game.playerIsOnTile()) {
            self.game.playerVelocity.y = std.math.min(maxFallVelocity, self.game.playerVelocity.y + gravity * dt);
        }

        // self.game.playerVelocity.y = std.math.min(maxFallVelocity, self.game.playerVelocity.y + self.game.playerAcceleration.y * dt);
        if (self.game.playerVelocity.y != 0) {
            self.game.playerFallingAnimation.tick(t);
            self.game.playerIdleAnimation.reset();
            self.game.playerWalkingAnimation.reset();
        } else if (self.game.playerVelocity.x != 0 and self.game.playerVelocity.y == 0) {
            self.game.playerWalkingAnimation.tick(t);
            self.game.playerFallingAnimation.tick(t);
            self.game.playerIdleAnimation.reset();
        } else if (self.game.playerVelocity.x == 0) {
            self.game.playerIdleAnimation.tick(t);
            self.game.playerFallingAnimation.tick(t);
            self.game.playerWalkingAnimation.reset();
        }
        self.game.movePlayer(dt);

        self.game.starAnimation.tick(t);
    }

    pub fn render(self: *GameScene, ctx: *Context) void {
        const renderer = ctx.renderer;
        const viewport = renderer.viewport;

        const bounds = viewport.bounds;
        const background = renderer.textures.get("background_jungle").?;
        const backgroundColor = allegro.mapRgb(211, 229, 252);
        const backgroundDisplacementFactor = -0.01;

        const backgroundCenter = self.game.playerPosition.x * backgroundDisplacementFactor;
        const backgroundLeft = @mod(backgroundCenter, 1.0) - 1;

        allegro.clearToColor(backgroundColor);

        background.drawScaled(bounds.min.x + backgroundLeft * bounds.width(), bounds.min.y, bounds.width(), bounds.height());
        background.drawScaled(bounds.min.x + (backgroundLeft + 1) * bounds.width(), bounds.min.y, bounds.width(), bounds.height());

        const tileBounds = self.game.tilesInView();

        const tiles = self.game.level.tiles;
        const collectables = self.game.level.collectables;

        var x = tileBounds.min.x;
        while (x < tileBounds.max.x) : (x += 1) {
            const randomDirtOffset = self.game.randomTileOffset(x, 99, 3);
            self.game.drawSpriteFrame("tiles_dirt_32", 0 + randomDirtOffset, x, 19);
            self.game.drawSpriteFrame("tiles_dirt_32", 3 + randomDirtOffset, x, 20);
        }

        var y = tileBounds.min.y;
        while (y <= tileBounds.max.y) : (y += 1) {
            x = tileBounds.min.x;
            while (x < tileBounds.max.x) : (x += 1) {
                if (tiles.get(x, y)) |tile| {
                    switch (tile) {
                        .Grass => {
                            const ordinals = tiles.getOrdinals(x, y, .Grass);
                            var offset: usize = 17;
                            switch (ordinals.count()) {
                                0 => offset = 16,
                                1 => {
                                    if (ordinals.right) offset = 12;
                                    if (ordinals.left) offset = 13;
                                    if (ordinals.top) offset = 15;
                                    if (ordinals.bottom) offset = 14;
                                },
                                2 => {
                                    if (ordinals.right and ordinals.left) offset = 0 + self.game.randomTileOffset(x, y, 4);
                                    if (ordinals.top and ordinals.bottom) offset = 20 + self.game.randomTileOffset(x, y, 2);
                                    if (ordinals.top and ordinals.left) offset = 9;
                                    if (ordinals.top and ordinals.right) offset = 8;
                                    if (ordinals.right and ordinals.bottom) offset = 4;
                                    if (ordinals.left and ordinals.bottom) offset = 5;
                                },
                                3 => {
                                    if (!ordinals.left) offset = 6;
                                    if (!ordinals.top) offset = 18 + self.game.randomTileOffset(x, y, 2);
                                    if (!ordinals.right) offset = 7;
                                    if (!ordinals.bottom) offset = 10 + self.game.randomTileOffset(x, y, 2);
                                },
                                else => {},
                            }
                            self.game.drawSpriteFrame("tiles_grass_32", offset, x, y - 1);
                        },
                        // else => {},
                    }
                }

                if (collectables.get(x, y)) |collectable| {
                    if (self.game.collected.hasValue(x, y, true)) continue;

                    switch (collectable) {
                        .Star => {
                            const distanceToPlayer = engine.pt(x, y).float().distance(self.game.playerPosition);
                            self.game.drawSpriteFrame("item_star_32", self.game.starAnimation.currentOffset(@floatToInt(usize, @mod(distanceToPlayer * 0.54, 1) * 16)), x, y);
                        },
                        .Exit => {
                            const position = self.game.tileP(engine.pt(x, y).float().subtract(engine.ptF(0.5, 1)));
                            renderer.drawTextureV("exit_portal", position.x, position.y);
                        },
                    }
                }
            }
        }

        const playerDirectionFrameOffset: usize = if (self.game.playerDirection == game.Game.Direction.Left) 0 else 12;
        if (self.game.playerVelocity.y != 0) {
            self.game.drawSpriteFrameP("character_lion_48", self.game.playerFallingAnimation.current + playerDirectionFrameOffset, self.game.playerPosition.add(engine.ptF(-0.25, -0.25)));
        } else if (self.game.playerVelocity.x != 0 and self.game.playerVelocity.y == 0) {
            self.game.drawSpriteFrameP("character_lion_48", self.game.playerWalkingAnimation.current + playerDirectionFrameOffset, self.game.playerPosition.add(engine.ptF(-0.25, -0.25)));
        } else {
            self.game.drawSpriteFrameP("character_lion_48", self.game.playerIdleAnimation.current + playerDirectionFrameOffset, self.game.playerPosition.add(engine.ptF(-0.25, -0.25)));
        }

        ctx.renderer.printTextV("default", ctx.palette.background.text, 0.01, 0.01, .Left, "Stars collected: {d}", .{self.game.starsCollected});
        const percentage = self.game.starsCollected * 100 / self.game.starsTotal;
        ctx.renderer.printTextV("default", ctx.palette.background.text, 0.96, 0.01, .Right, "{d} %", .{percentage});
        if (percentage > 0) {
            ctx.renderer.drawTintedSpriteFrameV("pie_charts_24", percentage - 1, ctx.palette.background.text, 0.97, 0.01);
        }

        const textColor = allegro.mapRgb(0x48, 0x91, 0x00);

        if (self.game.state != .InProgress) {
            ctx.renderer.textures.get("opaque").?.drawTintedScaled(allegro.mapRgba(0, 0, 0, 191), 0, 0, @intToFloat(f32, ctx.renderer.display.width()), @intToFloat(f32, ctx.renderer.display.height()));

            if (self.game.state == .Over) {
                ctx.renderer.drawTextV("default", textColor, 0.5, 0.2, .Center, "Game over");
            }
            if (self.game.state == .Completed) {
                ctx.renderer.drawTextV("default", textColor, 0.5, 0.2, .Center, "Level completed");
            }

            if (self.game.starsCollected < self.game.starsTotal) {
                ctx.renderer.drawSpriteFrameV("text_balloons", 0, 0.45, 0.40);
                ctx.renderer.drawTextV("small", ctx.palette.background.text, 0.517, 0.436, .Center, "Do you want to");
                ctx.renderer.drawTextV("small", ctx.palette.background.text, 0.517, 0.465, .Center, "try again?");
                ctx.renderer.drawSpriteFrameV("character_lion_48", self.game.playerIdleAnimation.current + 12, 0.41, 0.57);
                ctx.renderer.drawTextV("small", textColor, 0.5, 0.7, .Center, "[enter] try again | [escape] quit");
            }
        }

        if (ctx.showDebug) {
            ctx.renderer.printTextV("debug", ctx.palette.background.text, 0.01, 0.1, .Left, "Character: ({d:.2}, {d:.2})", .{ self.game.playerPosition.x, self.game.playerPosition.y });
            ctx.renderer.printTextV("debug", ctx.palette.background.text, 0.01, 0.15, .Left, "Tiles: ({d}, {d}) -> ({d}, {d})", .{ tileBounds.min.x, tileBounds.min.y, tileBounds.max.x, tileBounds.max.y });
            // self.game.debugHighlightTile(self.game.playerPosition);
        }
    }
};
