const std = @import("std");
const allegro = @import("allegro");
const engine = @import("engine.zig");
const game = @import("game.zig");
const paths = @import("paths.zig");
const Context = @import("context.zig").Context;
const GameScene = @import("game_scene.zig").GameScene;
const Palette = @import("palette.zig").Palette;
const Renderer = @import("renderer.zig").Renderer;
const TitleScene = @import("title_scene.zig").TitleScene;

fn hexColor(hex: []const u8) allegro.Color {
    return allegro.Color.initFromHex(hex);
}

pub fn main() !void {
    // Primary: #551485
    // Secondary: #e44b50
    // Tertiary: #479be6
    // Neutral: #948f94

    var arena = std.heap.ArenaAllocator.init(std.heap.c_allocator);
    defer arena.deinit();
    const allocator = arena.allocator();

    var context = try Context.init(allocator);
    defer context.deinit();
    context.palette = Palette{
        .primary = .{ .background = hexColor("#7d42ad"), .text = hexColor("#ffffff") },
        .primaryContainer = .{ .background = hexColor("#f2daff"), .text = hexColor("#2e004e") },
        .secondary = .{ .background = hexColor("#b32731"), .text = hexColor("#ffffff") },
        .secondaryContainer = .{ .background = hexColor("#ffdad8"), .text = hexColor("#410006") },
        .tertiary = .{ .background = hexColor("#0062a0"), .text = hexColor("#ffffff") },
        .tertiaryContainer = .{ .background = hexColor("#d0e4ff"), .text = hexColor("#001d35") },
        .background = .{ .background = hexColor("#fffbff"), .text = hexColor("#2b0052") },
        .outline = hexColor("#7c757e"),
    };
    const renderer = &context.renderer;

    allegro.setNewBitmapFlags(allegro.NewBitmapFlags{ .MIN_LINEAR = true, .MAG_LINEAR = true });
    try renderer.textures.addFromFile("opaque", paths.AssetsDir ++ "/images/opaque.png");
    try renderer.textures.addFromFile("title_untitled", paths.AssetsDir ++ "/images/title_untitled.png");
    try renderer.textures.addFromFile("background_jungle", paths.AssetsDir ++ "/images/background_jungle.png");

    allegro.setNewBitmapFlags(allegro.NewBitmapFlags{ .MIN_LINEAR = false, .MAG_LINEAR = false });
    try renderer.textures.addFromFile("character_lion_48", paths.AssetsDir ++ "/images/character_lion_48.png");
    try renderer.textures.addFromFile("exit_portal", paths.AssetsDir ++ "/images/exit_portal.png");
    try renderer.textures.addFromFile("item_star_32", paths.AssetsDir ++ "/images/item_star_32.png");
    try renderer.textures.addFromFile("pie_charts_24", paths.AssetsDir ++ "/images/pie_charts_24.png");
    try renderer.textures.addFromFile("text_balloons", paths.AssetsDir ++ "/images/text_balloons.png");
    try renderer.textures.addFromFile("tiles_dirt_32", paths.AssetsDir ++ "/images/tiles_dirt_32.png");
    try renderer.textures.addFromFile("tiles_grass_32", paths.AssetsDir ++ "/images/tiles_grass_32.png");

    try renderer.sprites.addFromTextures(renderer.textures, "character_lion_48", 48, 48);
    try renderer.sprites.addFromTextures(renderer.textures, "item_star_32", 32, 32);
    try renderer.sprites.addFromTextures(renderer.textures, "pie_charts_24", 24, 24);
    try renderer.sprites.addFromTextures(renderer.textures, "text_balloons", 128, 96);
    try renderer.sprites.addFromTextures(renderer.textures, "tiles_dirt_32", 32, 32);
    try renderer.sprites.addFromTextures(renderer.textures, "tiles_grass_32", 32, 64);

    allegro.convertMemoryBitmaps();

    try context.registerFonts();

    try context.switchToScene(GameScene, null);

    var t = allegro.getTime();
    var fpsBuffer = [_]u8{0} ** 32;
    _ = fpsBuffer;
    while (!context.shouldQuit) {
        const scene = if (context.scene) |scene| scene else {
            break;
        };

        const newT = allegro.getTime();
        const deltaT = @floatCast(f32, newT - t);
        t = newT;
        context.fps.update(deltaT);
        scene.tick(&context, @floatCast(f32, t), deltaT);

        while (!context.events.isEmpty()) {
            var event: allegro.Event = undefined;
            _ = context.events.get(&event);
            switch (event.type) {
                allegro.c.ALLEGRO_EVENT_DISPLAY_CLOSE => context.quit(),
                allegro.c.ALLEGRO_EVENT_DISPLAY_RESIZE => {
                    context.renderer.resized(event.display.width, event.display.height);
                    try context.registerFonts();
                    _ = allegro.acknowledgeResize(allegro.Display{ .native = event.display.source.? });
                },
                allegro.c.ALLEGRO_EVENT_KEY_CHAR => {
                    switch (event.keyboard.keycode) {
                        allegro.c.ALLEGRO_KEY_ESCAPE => context.quit(),
                        allegro.c.ALLEGRO_KEY_F9 => context.showFPS = !context.showFPS,
                        allegro.c.ALLEGRO_KEY_F10 => context.showDebug = !context.showDebug,
                        allegro.c.ALLEGRO_KEY_F11 => renderer.toggleFullScreen(),
                        allegro.c.ALLEGRO_KEY_F12 => try renderer.takeScreenshot(),
                        else => {},
                    }
                },
                allegro.c.ALLEGRO_EVENT_KEY_DOWN => {
                    try context.keys.press(event.keyboard.keycode);
                },
                allegro.c.ALLEGRO_EVENT_KEY_UP => {
                    try context.keys.release(event.keyboard.keycode);
                },
                else => {},
            }
            if (context.shouldQuit) {
                break;
            }
            try scene.handle(&context, &event);
        }

        scene.render(&context);
        if (context.showFPS) {
            renderer.printTextV("debug", context.palette.background.text, 0.99, 0.01, Renderer.TextAlignment.Right, "FPS: {}", .{context.fps.fps()});
        }

        allegro.flipDisplay();
    }
}

test "simple test" {
    var list = std.ArrayList(i32).init(std.testing.allocator);
    defer list.deinit(); // try commenting this out and see if zig detects the memory leak!
    try list.append(42);
    try std.testing.expectEqual(@as(i32, 42), list.pop());
}
