const std = @import("std");
const allegro = @import("allegro");
const engine = @import("engine.zig");
const Context = @import("context.zig").Context;

pub const Renderer = struct {
    textFormattingBuffer: [1024]u8 = [_]u8{0} ** 1024,

    display: allegro.Display,
    viewport: engine.Viewport,
    fonts: engine.Fonts,
    textures: engine.Textures,
    sprites: engine.Sprites,

    pub const TextAlignment = enum {
        Center,
        Left,
        Right,

        pub fn toDrawTextFlags(self: TextAlignment) allegro.DrawTextFlags {
            return switch (self) {
                .Center => allegro.DrawTextFlags.ALIGN_CENTER,
                .Left => allegro.DrawTextFlags.ALIGN_LEFT,
                .Right => allegro.DrawTextFlags.ALIGN_RIGHT,
            };
        }
    };

    pub fn init(allocator: std.mem.Allocator, display: allegro.Display) Renderer {
        const viewport = engine.Viewport.init(display.width(), display.height());

        return Renderer{
            .display = display,
            .viewport = viewport,
            .fonts = engine.Fonts.init(allocator),
            .textures = engine.Textures.init(allocator),
            .sprites = engine.Sprites.init(allocator),
        };
    }

    pub fn deinit(self: *Renderer) void {
        self.fonts.deinit();
        self.textures.deinit();
        self.sprites.deinit();
        self.display.destroy();
    }

    pub fn drawSpriteFrame(self: Renderer, spriteName: []const u8, frame: usize, x: f32, y: f32) void {
        if (self.sprites.getFrame(spriteName, frame)) |sprite| {
            sprite.drawScaledUniform(x, y, self.viewport.scale);
        }
    }

    pub fn drawSpriteFrameV(self: Renderer, spriteName: []const u8, frame: usize, x: f32, y: f32) void {
        const screen = self.viewport.viewToScreen(x, y);
        self.drawSpriteFrame(spriteName, frame, screen.x, screen.y);
    }

    pub fn drawText(self: Renderer, fontName: []const u8, color: allegro.Color, x: f32, y: f32, alignment: TextAlignment, text: [*:0]const u8) void {
        if (self.fonts.get(fontName)) |font| {
            font.draw(color, x, y, alignment.toDrawTextFlags(), text);
        }
    }

    pub fn drawTexture(self: Renderer, textureName: []const u8, x: f32, y: f32) void {
        if (self.textures.get(textureName)) |texture| {
            texture.drawScaledUniform(x, y, self.viewport.scale);
        }
    }

    pub fn drawTextureV(self: Renderer, textureName: []const u8, x: f32, y: f32) void {
        const screen = self.viewport.viewToScreen(x, y);
        self.drawTexture(textureName, screen.x, screen.y);
    }

    pub fn drawTextV(self: Renderer, fontName: []const u8, color: allegro.Color, x: f32, y: f32, alignment: TextAlignment, text: [*:0]const u8) void {
        const screen = self.viewport.viewToScreen(x, y);
        self.drawText(fontName, color, screen.x, screen.y, alignment, text);
    }

    pub fn drawTintedSpriteFrame(self: Renderer, spriteName: []const u8, frame: usize, tint: allegro.Color, x: f32, y: f32) void {
        if (self.sprites.getFrame(spriteName, frame)) |sprite| {
            sprite.drawTintedScaledUniform(tint, x, y, self.viewport.scale);
        }
    }

    pub fn drawTintedSpriteFrameV(self: Renderer, spriteName: []const u8, frame: usize, tint: allegro.Color, x: f32, y: f32) void {
        const screen = self.viewport.viewToScreen(x, y);
        if (self.sprites.getFrame(spriteName, frame)) |sprite| {
            sprite.drawTintedScaledUniform(tint, screen.x, screen.y, self.viewport.scale);
        }
    }

    pub fn printText(self: *Renderer, font: []const u8, color: allegro.Color, x: f32, y: f32, alignment: TextAlignment, comptime fmt: []const u8, args: anytype) void {
        const text = std.fmt.bufPrintZ(&self.textFormattingBuffer, fmt, args) catch {
            return;
        };
        self.drawText(font, color, x, y, alignment, text);
    }

    pub fn printTextV(self: *Renderer, font: []const u8, color: allegro.Color, x: f32, y: f32, alignment: TextAlignment, comptime fmt: []const u8, args: anytype) void {
        const screen = self.viewport.viewToScreen(x, y).round();
        self.printText(font, color, screen.x, screen.y, alignment, fmt, args);
    }

    pub fn resized(self: *Renderer, width: i32, height: i32) void {
        self.viewport.update(width, height);
    }

    pub fn takeScreenshot(self: Renderer) !void {
        const screen = try allegro.getBackbuffer(self.display);
        _ = allegro.saveBitmap("screenshot.png", screen);
    }

    pub fn toggleFullScreen(self: *Renderer) void {
        var displayFlags = allegro.getDisplayFlags(self.display);
        _ = allegro.setDisplayFlag(self.display, allegro.NewDisplayFlags{ .FULLSCREEN_WINDOW = true }, !displayFlags.FULLSCREEN_WINDOW);
        self.viewport.update(self.display.width(), self.display.height());
    }
};
